#include <iostream>

#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif

#ifdef __unix__
#include <GL/glut.h>
#endif

#ifdef _WIN32
#pragma comment(lib, "glew32.lib")

#include <GL\glew.h>
#include <GL\glut.h>
#endif

#include "CMeshLoaderSimple.h"
#include <stdlib.h>
#include <stdio.h>
#include <cmath>
#include "GLSL_helper.h"
#include "GeometryCreator.h"
#include "MStackHelp.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "glm/gtc/type_ptr.hpp"

using namespace std;
using namespace glm;

#define PI 3.14
#define EYE_LEVEL 1.25

GLuint triBuffObj;
GLuint colBuffObj;
GLuint normalBuffObj;

GLuint GrndBuffObj;
GLuint GIndxBuffObj;

unsigned int const MeshCount = 20;
Mesh *Meshes[MeshCount];
unsigned int CurrentMesh;

RenderingHelper ModelTrans;

int shade = 1;
int ShadeProg;
int TriangleCount;

int g_GiboLen;
int meshCount = 0;
float g_width;
float g_height;

GLint h_aPosition;
GLint h_aNormal;
GLint h_uModelMatrix;
GLint h_uViewMatrix;
GLint h_uProjMatrix;
GLint h_uLightPos;
GLint h_uLightColor;
GLint h_uCamPos;
GLint h_uMatAmb;
GLint h_uMatDif;
GLint h_uMatSpec;
GLint h_uMatShine;

float g_groundY = -0.5;
float g_groundSize = 15.0;

int winSize = 400;
int invertPitch = 1;

float g_mouseStart[] = {winSize/2, winSize/2}, g_mouseEnd[] = {winSize/2, winSize/2}, g_mouseDiff[] = {0, 0};

float alpha = 0;
float beta = -PI / 2; // -PI/2 to start facing in front direction without jumping

glm::vec3 eyePos = glm::vec3(0, 0, 0); // camera position, but also corresponds to player position
glm::vec3 lookAtPt = glm::vec3(0, 0, 0);
vec3 gazeVector = vec3(0, 0, 0);
vec3 wVector = vec3(0, 0, -1);
vec3 uVector = vec3(-1, 0, 0);
vec3 vVector = vec3(0, 1, 0);
vec3 upVector = vec3(0, 1, 0);
vec4 lightVector = vec4(0);

//    GAME LOGIC FLAGS
int g_hasKey = false;
int g_chairPulled = false; // if true, player can access laptop on table
int g_doorOpened = 0; // if true, player wins
int g_laptopClosed = false; // if true, player can phone for help
int g_phonedForHelp = false; // if true, key will fly through window
int g_powerIsOn = false; // if true, player can use laptop
//

//    GAME LOGIC GLOBALS
vec3 g_doorPos = vec3(-1, 1, -6);
float g_distToDoor = 0;
float g_doorAngle = 0;

vec3 g_chairPos = vec3(-4.5, 0, -4.5);
float g_chairTransZ = -4.5;
float g_distToChair = 0;

vec3 g_laptopPos = vec3(-4.5, 0.7, -4.6);
float g_laptopAngle = 90;
float g_distToLaptop = 0;

vec3 g_fuseBoxPos = vec3(5.8, 1.5, -2.5);
float g_fuseBoxAngle = 0;
float g_distToFuseBox = 0;

vec3 g_phonePos = vec3(5.3, 1, 0);
float g_distToPhone = 0;

vec3 g_planePos = vec3(2.8, 0.5, -8);
float g_distToPlane = 0;


vec3 diff = vec3(0, 0, 0);
//

//	  STRINGS
char doorText[] = "Hold 'e' near door to open the door.";
char chairText[] = "Hold 'e' near chair to pull out chair.";
char laptopText[] = "Hold 'e' near laptop to search for rescue hotline number.";
char fuseBoxText[] = "Hold 'e' near fuse box to open it and turn power on.";
char phoneText[] = "Press 'e' near phone to phone for help.";
char planeText[] = "Press 'e' near rescue plane to take the key.";

//

float lightX = 0, lightY = 6, lightZ = 0;

// TEXTURE STUFF

/*data structure for the image used for  texture mapping */
typedef struct Image {
	unsigned long sizeX;
	unsigned long sizeY;
	char *data;
} Image;

Image *TextureImage;

typedef struct RGB {
	GLubyte r;
	GLubyte g;
	GLubyte b;
} RGB;

RGB myimage[64][64];
RGB* g_pixel;

//forward declaration of image loading and texture set-up code
int ImageLoad(char *filename, Image *image);
GLvoid LoadTexture(char* image_file, int tex_id);

GLint h_uTexUnit;
GLint h_aTexCoord;

GLuint TexBuffObj;
GLuint GTexBuffObj;
GLuint SkyTexBuffObj;


GLint h_texFlag;
GLint h_roomFlag;
GLint h_texOnlyFlag;
GLint h_gameWinFlag;

float roomMinZ = -5.0, roomMaxZ = 5.0, roomMinX = -5.0, roomMaxX = 5.0;

//

int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName) {
  GLuint VS; //handles to shader object
  GLuint FS; //handles to frag shader object
  GLint vCompiled, fCompiled, linked; //status of shader

  VS = glCreateShader(GL_VERTEX_SHADER);
  FS = glCreateShader(GL_FRAGMENT_SHADER);

  //load the source
  glShaderSource(VS, 1, &vShaderName, NULL);
  glShaderSource(FS, 1, &fShaderName, NULL);

  //compile shader and print log
  glCompileShader(VS);
  /* check shader status requires helper functions */
  printOpenGLError();
  glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
  printShaderInfoLog(VS);

  //compile shader and print log
  glCompileShader(FS);
  /* check shader status requires helper functions */
  printOpenGLError();
  glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
  printShaderInfoLog(FS);

  if (!vCompiled || !fCompiled) {
    printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
    return 0;
  }

  //create a program object and attach the compiled shader
  ShadeProg = glCreateProgram();
  glAttachShader(ShadeProg, VS);
  glAttachShader(ShadeProg, FS);

  glLinkProgram(ShadeProg);
  /* check shader status requires helper functions */
  printOpenGLError();
  glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
  printProgramInfoLog(ShadeProg);

  glUseProgram(ShadeProg);

  /* get handles to attribute data */
  h_uLightPos = safe_glGetUniformLocation(ShadeProg, "uLightPos");
  h_uLightColor = safe_glGetUniformLocation(ShadeProg, "uLightColor");
  h_uCamPos = safe_glGetUniformLocation(ShadeProg, "uCamPos");

  h_aPosition = safe_glGetAttribLocation(ShadeProg, "aPosition");
  h_aNormal = safe_glGetAttribLocation(ShadeProg, "aNormal");
  h_uProjMatrix = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
  h_uViewMatrix = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
  h_uModelMatrix = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
  h_uMatAmb = safe_glGetUniformLocation(ShadeProg, "uMat.aColor");
  h_uMatDif = safe_glGetUniformLocation(ShadeProg, "uMat.dColor");
  h_uMatSpec = safe_glGetUniformLocation(ShadeProg, "uMat.sColor");
  h_uMatShine = safe_glGetUniformLocation(ShadeProg, "uMat.shine");

  h_aTexCoord = safe_glGetAttribLocation(ShadeProg,  "aTexCoord");
  h_uTexUnit = safe_glGetUniformLocation(ShadeProg, "uTexUnit");
  h_texFlag = safe_glGetUniformLocation(ShadeProg, "texFlag");
  h_roomFlag = safe_glGetUniformLocation(ShadeProg, "roomFlag");
  h_texOnlyFlag = safe_glGetUniformLocation(ShadeProg, "texOnlyFlag");
  h_gameWinFlag = safe_glGetUniformLocation(ShadeProg, "gameWinFlag");

  return 1;
}

void SetProjectionMatrix(int usePerspective) {

	mat4 Projection = perspective(90.0f, (float)g_width/g_height, 0.1f, 100.f);

	safe_glUniformMatrix4fv(h_uProjMatrix, value_ptr(Projection));
}


void SetView() {
  lookAtPt.x = 1 * cos(alpha) * cos(beta);
  lookAtPt.y = invertPitch * sin(alpha);
  lookAtPt.z = 1 * cos(alpha) * cos(PI / 2 - beta);

  lookAtPt *= winSize; // multiply by window size

  gazeVector = lookAtPt - eyePos;
  wVector = -normalize(gazeVector);
  uVector = normalize(cross(upVector, wVector));
  vVector = normalize(cross(uVector, wVector));

  eyePos.y = EYE_LEVEL; // to prevent clipping through the ground

  glm::mat4 LookAtView = glm::lookAt(eyePos, lookAtPt, upVector);
  safe_glUniformMatrix4fv(h_uViewMatrix, glm::value_ptr(LookAtView));
}

void SetModel() {
//  if (lightVector == vec4(0)) {
//    lightVector = ModelTrans.modelViewMatrix * vec4(0, 0, -2, 1);
//    lightVector = vec4(15, 18, 0, 1) - lightVector;
//  }
  lightVector = vec4(lightX, lightY, lightZ, 1);
  safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(ModelTrans.modelViewMatrix));
}

void SetMaterial(int i) {
  glUseProgram(ShadeProg);
  switch (i) {
    case 1: // shiny green/jade
      safe_glUniform3f(h_uMatAmb, 0.2, 0.2, 0.2);
      safe_glUniform3f(h_uMatDif, 0.0, 0.5, 0.08);
      safe_glUniform3f(h_uMatSpec, 0.4, 0.8, 0.8);
      safe_glUniform1f(h_uMatShine, 200.0);
      break;
    case 2: // room cube coloring
      safe_glUniform3f(h_uMatAmb, 0.5, 0.1, 0.1);
      safe_glUniform3f(h_uMatDif, 0.4, 0.1, 0.1);
      safe_glUniform3f(h_uMatSpec, 0.5, 0.0, 0.0);
      safe_glUniform1f(h_uMatShine, 1.0);
      break;
    case 3: // plane coloring
      safe_glUniform3f(h_uMatAmb, 0.2, 0.2, 0.2);
      safe_glUniform3f(h_uMatDif, 0.5, 0.0, 0.08);
      safe_glUniform3f(h_uMatSpec, 0.6, 0.3, 0.4);
      safe_glUniform1f(h_uMatShine, 10.0);
      break;
    case 4: // shiny blue
      safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.3);
      safe_glUniform3f(h_uMatDif, 0.3, 0.3, 0.5);
      safe_glUniform3f(h_uMatSpec, 0.2, 0.2, 0.2);
      safe_glUniform1f(h_uMatShine, 200.0);
      break;
    case 5: // light brown wood
      safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.1);
      safe_glUniform3f(h_uMatDif, 0.5, 0.2, 0.0);
      safe_glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1);
      safe_glUniform1f(h_uMatShine, 2.0);
      break;
    case 6: // shiny yellow
      safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.1);
      safe_glUniform3f(h_uMatDif, 0.8, 0.6, 0.0);
      safe_glUniform3f(h_uMatSpec, 0.8, 0.6, 0.1);
      safe_glUniform1f(h_uMatShine, 200.0);
      break;
    case 7: // matte grey
      safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.1);
      safe_glUniform3f(h_uMatDif, 0.1, 0.1, 0.1);
      safe_glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1);
      safe_glUniform1f(h_uMatShine, 20.0);
      break;
  }
}

/* intialize the texture data */
static void initTexture() {

	 static GLfloat CubeTex[] = {
			0, 0, // back
			0, 1,
			1, 1,
			1, 0

		};

	 static GLfloat GCubeTex[] = {
	 			0, 0, // back
	 			0, 0,
	 			0, 0,
	 			0, 0,

	 			0, 0, // front
	 			0, 0,
	 			0, 0,
	 			0, 0,

	 			0, 0, // left
	 			0, 0,
	 			0, 0,
	 			0, 0,

	 			0, 0, // right
	 			0, 0,
	 			0, 0,
	 			0, 0,

	 			0, 0, // top
	 			0, 1,
	 			1, 1,
	 			1, 0
	 };

	 static GLfloat SkyCubeTex[] = {
	 	 			0, 0, // back
	 	 			0, 0.5,
	 	 			0.5, 0.5,
	 	 			0.5, 0,

	 	 			0, 0, // front
	 	 			0, 0.5,
	 	 			0.5, 0.5,
	 	 			0.5, 0,

	 	 			0, 0, // left
	 	 			0, 0.5,
	 	 			0.5, 0.5,
	 	 			0.5, 0,

	 	 			0, 0, // right
	 	 			0, 0.5,
	 	 			0.5, 0.5,
	 	 			0.5, 0,

	 	 			0.5, 0, // top
	 	 			0.5, .7,
	 	 			1, .7,
	 	 			1, 0
	 	 };

		glGenBuffers(1, &TexBuffObj);
		glBindBuffer(GL_ARRAY_BUFFER, TexBuffObj);
		glBufferData(GL_ARRAY_BUFFER, sizeof(CubeTex), CubeTex, GL_STATIC_DRAW);

		glGenBuffers(1, &GTexBuffObj);
		glBindBuffer(GL_ARRAY_BUFFER, GTexBuffObj);
		glBufferData(GL_ARRAY_BUFFER, sizeof(GCubeTex), GCubeTex, GL_STATIC_DRAW);

		glGenBuffers(1, &SkyTexBuffObj);
		glBindBuffer(GL_ARRAY_BUFFER, SkyTexBuffObj);
		glBufferData(GL_ARRAY_BUFFER, sizeof(SkyCubeTex), SkyCubeTex, GL_STATIC_DRAW);

}

void InitGeom() {
  CMeshLoader::loadVertexBufferObjectFromMesh("Models/cessna_color500.m", TriangleCount, triBuffObj, colBuffObj, normalBuffObj);
//  initGround();
  initTexture();

}

void Initialize() {
	glClearColor(0, 0, 0, 1.0f);
	glEnable(GL_DEPTH_TEST);
  ModelTrans.useModelViewMatrix();
  ModelTrans.loadIdentity();

  /* texture specific settings */
  glEnable(GL_TEXTURE_2D);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

void updatePitchAndYaw() {
   g_mouseDiff[0] = g_mouseEnd[0] - g_mouseStart[0]; // x difference
   g_mouseDiff[1] = g_mouseEnd[1] - g_mouseStart[1]; // y difference

   if (alpha + g_mouseDiff[1] * -PI / g_height < 1.3962634
    && alpha + g_mouseDiff[1] * -PI / g_height > -1.3962634) { // restrict vertical movement to 180 degrees total
      alpha += g_mouseDiff[1] * -PI / g_height;
   }

   beta += g_mouseDiff[0] * PI / g_width;

   g_mouseStart[0] = g_mouseEnd[0]; // update x and y to last position
   g_mouseStart[1] = g_mouseEnd[1];
}

void drawPlane(vec3 translation) {
	ModelTrans.pushMatrix();
	   ModelTrans.translate(translation);
	   ModelTrans.rotate(90, vec3(0, 1, 0));

	   SetModel();
	   	SetMaterial(4);
	   	safe_glEnableVertexAttribArray(h_aPosition);
	   	glBindBuffer(GL_ARRAY_BUFFER, triBuffObj);
	   	safe_glVertexAttribPointer(h_aPosition, 4, GL_FLOAT, GL_FALSE, 0, 0);

	   	safe_glEnableVertexAttribArray(h_aNormal);
	   	glBindBuffer(GL_ARRAY_BUFFER, normalBuffObj);
	   	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
	   glDrawArrays(GL_TRIANGLES, 0, TriangleCount * 3);

	ModelTrans.popMatrix();
}

void drawMesh(int meshNum, int matNum) {
	SetModel();
	SetMaterial(matNum);
	safe_glEnableVertexAttribArray(h_aPosition);
	glBindBuffer(GL_ARRAY_BUFFER, Meshes[meshNum]->PositionHandle);
	safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

	safe_glEnableVertexAttribArray(h_aNormal);
	glBindBuffer(GL_ARRAY_BUFFER, Meshes[meshNum]->NormalHandle);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, Meshes[meshNum]->IndexHandle);
	glDrawElements(GL_TRIANGLES, Meshes[meshNum]->IndexBufferLength, GL_UNSIGNED_SHORT, 0);
}

void drawCubeBase() {
  SetModel();
  drawMesh(1, 2);
}

void drawRoomCube() {
	ModelTrans.loadIdentity();
	ModelTrans.pushMatrix();
	   drawMesh(0, 2);
	ModelTrans.popMatrix();
}

void drawTable(float Tx, float Ty, float Tz, vec3 scale) {
	ModelTrans.pushMatrix();
      ModelTrans.translate(vec3(Tx, Ty, Tz)); // translate table
      ModelTrans.scale(scale.x, scale.y, scale.z);

      ModelTrans.pushMatrix(); // leg 1
         // translate and draw leg
         ModelTrans.translate(vec3(-1, -1, -1));
         drawMesh(2, 5);
      ModelTrans.popMatrix(); // return to tabletop frame

      ModelTrans.pushMatrix(); // leg 2
         ModelTrans.translate(vec3(-1, -1, 1));
         drawMesh(2, 5);
      ModelTrans.popMatrix(); // return to tabletop frame

      ModelTrans.pushMatrix(); // leg 3
         ModelTrans.translate(vec3(1, -1, 1));
         drawMesh(2, 5);
      ModelTrans.popMatrix(); // return to tabletop frame

      ModelTrans.pushMatrix(); // leg 4
         ModelTrans.translate(vec3(1, -1, -1));
         drawMesh(2, 5);
      ModelTrans.popMatrix(); // return to tabletop frame

      // draw tabletop
      drawMesh(1, 5);

   ModelTrans.popMatrix();
}

void drawDoor(vec3 translation) {
	ModelTrans.pushMatrix();
      ModelTrans.translate(translation); // translate door to final position
      ModelTrans.rotate(360.0 - g_doorAngle, vec3(0, 1, 0)); // rotate door around hinge
      ModelTrans.translate(vec3(0.875, 0, 0)); // align door's left edge on origin


      ModelTrans.pushMatrix();
         // translate and draw front knob
         ModelTrans.translate(vec3(0.55, -0.1, 0.2));
         drawMesh(4, 6);
      ModelTrans.popMatrix();

      ModelTrans.pushMatrix();
         // translate and draw back knob
         ModelTrans.translate(vec3(0.55, -0.1, -0.2));
         drawMesh(4, 6);
      ModelTrans.popMatrix();

      // draw door panel
      drawMesh(3, 5);

   ModelTrans.popMatrix();
}

void drawChair(vec3 trans) {
	int material = 5;
	if (g_chairPulled) {
		material = 7;
	}

	ModelTrans.pushMatrix();
      ModelTrans.translate(vec3(trans.x, trans.y, g_chairTransZ)); // translate chair
      ModelTrans.scale(1, 1, 1);

      ModelTrans.pushMatrix(); // chair backing
         // translate and draw leg
         ModelTrans.translate(vec3(0, 0.5, 0.45));
         drawMesh(6, material);
      ModelTrans.popMatrix();

      ModelTrans.pushMatrix(); // leg 1
         // translate and draw leg
         ModelTrans.translate(vec3(-0.35, -1, -0.35));
         drawMesh(2, material);
      ModelTrans.popMatrix();

      ModelTrans.pushMatrix(); // back leg 2
         ModelTrans.translate(vec3(-0.35, -1, 0.37));
         drawMesh(2, material);
      ModelTrans.popMatrix();

      ModelTrans.pushMatrix(); // back leg 3
         ModelTrans.translate(vec3(0.35, -1, 0.37));
         drawMesh(2, material);
      ModelTrans.popMatrix();

      ModelTrans.pushMatrix(); // leg 4
         ModelTrans.translate(vec3(0.35, -1, -0.35));
         drawMesh(2, material);
      ModelTrans.popMatrix();

      // draw chair seat
      drawMesh(5, material);

   ModelTrans.popMatrix();
}

void drawLaptop(vec3 translation) {
	ModelTrans.pushMatrix();
      ModelTrans.translate(translation); // translate laptop to final position
      ModelTrans.scale(0.8, 0.6, 0.7);

      ModelTrans.pushMatrix();
         // translate and draw LCD screen in base's frame
         ModelTrans.translate(vec3(0, 0.055, -0.48));
         ModelTrans.rotate(360.0 - g_laptopAngle, vec3(1, 0, 0)); // rotate laptop around hinge
         ModelTrans.translate(vec3(0, 0, 0.5)); // align LCD's hinge on origin

         if (g_powerIsOn && !g_laptopClosed) {
        	 drawMesh(5, 4);
         }
         else {
        	 drawMesh(5, 7);
         }
      ModelTrans.popMatrix();

      // draw laptop base
      drawMesh(5, 7);

   ModelTrans.popMatrix();
}

void drawFuseBox(vec3 translation) {
	ModelTrans.pushMatrix();
      ModelTrans.translate(translation); // translate fusebox to final position
      ModelTrans.rotate(-90.0, vec3(0, 1, 0));
      ModelTrans.scale(0.8, 1, 1);

      ModelTrans.pushMatrix();
         // translate and draw fusebox cover
         ModelTrans.translate(vec3(-0.45, 0, 0.055));
         ModelTrans.rotate(360.0 - g_fuseBoxAngle, vec3(0, 1, 0)); // rotate fusebox cover around hinge
         ModelTrans.translate(vec3(0.5, 0, 0)); // align fusebox cover's hinge on origin
         ModelTrans.scale(1, 1, 0.5); // make the cover a little thinner

         drawMesh(6, 7);
      ModelTrans.popMatrix();

      // draw fuseBox base
      if (g_powerIsOn) { // if power is on, base of box turns green
    	  drawMesh(6, 1);
      }
      else {
    	  drawMesh(6, 7);
      }

   ModelTrans.popMatrix();
}

void drawPhone(vec3 translation) {
	int phoneMaterial = 7;

	if (g_laptopClosed) {
		phoneMaterial = 1;
	}

	ModelTrans.pushMatrix();
      ModelTrans.translate(translation); // translate phone to final position
      ModelTrans.rotate(-90.0, vec3(0, 1, 0));

      if (!g_phonedForHelp) {
    	  ModelTrans.pushMatrix();
    	  // translate and draw phone speaker
    	    ModelTrans.translate(vec3(0, 0.23, 0.05)); // align phone speaker on top

    	  drawMesh(8, phoneMaterial);
    	  ModelTrans.popMatrix();

    	  ModelTrans.pushMatrix();
    	  // translate and draw phone mouthpiece
    	    ModelTrans.translate(vec3(0, -0.23, 0.05)); // align mouthpiece on bottom

    	    drawMesh(8, phoneMaterial);
    	  ModelTrans.popMatrix();
      }


      ModelTrans.pushMatrix();
         drawTable(0, -0.48, 0, vec3(0.35, 0.65, 0.35)); // align night table below phone
      ModelTrans.popMatrix();

      // draw phone base
      if (!g_phonedForHelp) {
         drawMesh(7, phoneMaterial);
      }


   ModelTrans.popMatrix();
}

void drawRoom() {
	LoadTexture((char*)"roomAlpha.bmp", 0);

	glEnable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, 0);

	safe_glUniform1i(h_uTexUnit, 0);
	safe_glEnableVertexAttribArray(h_aTexCoord);
	glBindBuffer(GL_ARRAY_BUFFER, TexBuffObj);
	safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);

	safe_glUniform1i(h_texFlag, 1);
	safe_glUniform1i(h_roomFlag, 1);
	drawRoomCube();
	safe_glUniform1i(h_texFlag, 0);
	safe_glUniform1i(h_roomFlag, 0);
}

void drawGround() {
	glEnable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, 1);

	safe_glUniform1i(h_uTexUnit, 1);
	safe_glEnableVertexAttribArray(h_aTexCoord);
	glBindBuffer(GL_ARRAY_BUFFER, GTexBuffObj);
	safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);

	safe_glUniform1i(h_texFlag, 1);
	ModelTrans.pushMatrix();
	  ModelTrans.translate(vec3(0, -2.5, 0));
	  ModelTrans.scale(8, 8, 8);
	  drawMesh(1, 4); // draw ground
	ModelTrans.popMatrix();
	safe_glUniform1i(h_texFlag, 0);
}

void drawSkyBox() {
	glEnable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, 2);

	safe_glUniform1i(h_uTexUnit, 2);
	safe_glEnableVertexAttribArray(h_aTexCoord);
	glBindBuffer(GL_ARRAY_BUFFER, SkyTexBuffObj);
	safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);

	safe_glUniform1i(h_texOnlyFlag, 1);
	ModelTrans.pushMatrix();
	  ModelTrans.translate(vec3(0, 5, 0));
	  ModelTrans.scale(10, 30, 10);
	  drawMesh(1, 4); // draw skybox
	ModelTrans.popMatrix();
	safe_glUniform1i(h_texOnlyFlag, 0);
}

void Draw() {
  updatePitchAndYaw();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glUseProgram(ShadeProg);
  SetProjectionMatrix(1);
  SetView();
  SetModel();

  glUniform4f(h_uLightPos, lightVector.x, lightVector.y, lightVector.z, 1);
  glUniform3f(h_uLightColor, 1, 1, 1);
  glUniform4f(h_uCamPos, eyePos.x, eyePos.y, eyePos.z, 1);


  glUniform1i(h_gameWinFlag, g_doorOpened); // triggers end of game if door is opened
  if (g_doorOpened) {
	  glutSetWindowTitle("You Win! You've escaped the virtual room...now go outside in the real world. Press 'q' to quit.");
  }


  // DRAW MODELS
//  drawTable(lightX, lightY, lightZ, vec3(0.1, 0.1, 0.1));

  // Room
  drawRoom();

  if (!g_hasKey) {
     drawPlane(g_planePos);
  }

  drawTable(-4.5, 0.5, -5, vec3(1, 0.5, 0.5));
  drawDoor(g_doorPos);
  drawChair(g_chairPos);
  drawLaptop(g_laptopPos);
  drawFuseBox(g_fuseBoxPos);
  drawPhone(g_phonePos);

  // Ground
  drawGround();

  // SKYBOX
  drawSkyBox();

  // Disable the shader and attributes
  safe_glDisableVertexAttribArray(h_aPosition);
  safe_glDisableVertexAttribArray(h_aNormal);
  glUseProgram(0);

  /**
   * Display prompts to interact with objects when player is close to objects
   * and objects have not been used to their full potentials.
   */


  diff = g_doorPos - eyePos; // DOOR
  g_distToDoor = sqrtf(dot(diff, diff));
  if (g_hasKey && !g_doorOpened && g_distToDoor > 0.5 && g_distToDoor < 3.2) {
//	  drawText(-0.5, 0, 0.0006, 0.0006, doorText); // draw text
	  glutSetWindowTitle(doorText);
  }

  diff = g_chairPos - eyePos; // CHAIR
  g_distToChair = sqrtf(dot(diff, diff));
  if (!g_chairPulled && g_distToChair < 3.0) { // no prompt if pulled out completely
	  glutSetWindowTitle(chairText);
  }

  diff = g_laptopPos - eyePos; // LAPTOP
  g_distToLaptop = sqrtf(dot(diff, diff));
  if (g_powerIsOn && g_chairPulled && g_distToLaptop > 0.5 && g_distToLaptop < 2.25 && g_laptopAngle > 0) {
	  // only prompt if chair has been pulled out completely
	  glutSetWindowTitle(laptopText);
  }

  diff = g_fuseBoxPos - eyePos; // FUSEBOX
  g_distToFuseBox = sqrtf(dot(diff, diff));
  if (g_distToFuseBox > 0.5 && g_distToFuseBox < 2.25 && g_fuseBoxAngle < 130) {
	  // only prompt if chair has been pulled out completely
	  glutSetWindowTitle(fuseBoxText);
  }

  diff = g_phonePos - eyePos; // PHONE
  g_distToPhone = sqrtf(dot(diff, diff));
  if (g_laptopClosed && !g_phonedForHelp && g_distToPhone > 0.5 && g_distToPhone < 2.25) {
	  // only prompt if laptop has been closed completely
	  glutSetWindowTitle(phoneText);
  }

  diff = g_planePos - eyePos; // PHONE
  g_distToPlane = sqrtf(dot(diff, diff));
  if (!g_hasKey && g_phonedForHelp && g_distToPlane > 0.5 && g_distToPlane < 2.25) {
	  // only prompt if laptop has been closed completely
	  glutSetWindowTitle(planeText);
  }

  glutSwapBuffers();

  // DEBUGGING
  //printf("Player is at (%f, %f, %f)\n", eyePos.x, eyePos.y, eyePos.z);
}

void ReshapeGL(int width, int height) {
	g_width = (float)width;
	g_height = (float)height;
	glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));
}

void mouseHandler(int button, int state, int x, int y) {
  switch (state) {
    case GLUT_DOWN:
      g_mouseStart[0] = x;
      g_mouseStart[1] = y;
      break;
    case GLUT_UP:
      g_mouseEnd[0] = x;
      g_mouseEnd[1] = y;
      break;
  }
}

void motionHandler(int x, int y) {
  g_mouseEnd[0] = x;
  g_mouseEnd[1] = y;
  glutPostRedisplay();
}

// plane starts at 2.8, 0.5, -8
void animation() {
   if (g_phonedForHelp) {
	   if (g_planePos.y < 1.5 && g_planePos.z < -3) {
		   g_planePos.y += 0.045;
	   }
	   else if (g_planePos.y >= 1.5 && g_planePos.z < -3) {
		   g_planePos.z += 0.045;
	   }
	   else if (g_planePos.z >= -3 && g_planePos.y > 0) {
		   g_planePos.y -= 0.045;
	   }
   }

   glutPostRedisplay();
}

int isWithinRoom() {
   return (eyePos.z > roomMinZ && eyePos.z < roomMaxZ && eyePos.x > roomMinX && eyePos.x < roomMaxX);
}

void keyboard(unsigned char key, int x, int y ) {
  switch (key) {
    case 'w':
      wVector *= 0.2;
      if (g_doorOpened || (!g_doorOpened && isWithinRoom())) {
         eyePos -= wVector;
      }
      else {
    	 eyePos += wVector *= 2;
      }
      break;
    case 's':
      wVector *= 0.2;
      if (g_doorOpened || (!g_doorOpened && isWithinRoom())) {
    	  eyePos += wVector;
      }
      else {
    	  eyePos -= wVector *= 2;
      }
      break;
    case 'a':
      uVector *= 0.2;
      if (g_doorOpened || (!g_doorOpened && isWithinRoom())) {
    	  eyePos -= uVector;
      }
      else {
    	  eyePos += uVector *= 2;
      }
      break;
    case 'd':
    	uVector *= 0.2;
    	if (g_doorOpened || (!g_doorOpened && isWithinRoom())) {
    		eyePos += uVector;
    	}
    	else {
    		eyePos -= uVector *= 2;
    	}
      break;
    case 'q': case 'Q' :
      exit( EXIT_SUCCESS );
    case 'i':
      invertPitch *= -1;
      break;
//    case 'o': // hold to open door if player has key
//      if (g_hasKey && !g_doorOpened && g_distToDoor > 1.0 && g_distToDoor < 2.25  && g_doorAngle < 130.0) {
//         g_doorAngle += 5;
//      }
//      else if (g_doorAngle >= 130.0){
//    	  g_doorOpened = true;
//      }
//      break;
//    case 'c': // hold to pull out chair
//      if (!g_chairPulled && g_distToChair < 3.0 && g_chairTransZ < -2) {
//    	 g_chairTransZ += 0.1;
//      }
//      else if (g_chairTransZ >= -2){
//         g_chairPulled = true;
//      }
//      break;
//    case 'l': // hold to open laptop
//      if (!g_laptopClosed && g_distToLaptop > 1.0 && g_distToLaptop < 2.25 && g_laptopAngle > 0) {
//    	  g_laptopAngle -= 5;
//      }
//      else if (g_laptopAngle <= 0) {
//    	  g_laptopClosed = true;
//      }
//      break;
    case 'e': // action key
      // Can turn on power if close enough
      if (g_distToFuseBox > 1.0 && g_distToFuseBox < 2.25 && g_fuseBoxAngle < 130) {
          g_fuseBoxAngle += 5;
      }
      else if (g_fuseBoxAngle >= 130) {
    	  g_powerIsOn = true;
      }

      // Can pull out chair if close enough
      if (!g_chairPulled && g_distToChair < 3.0 && g_chairTransZ < -2) {
    	  g_chairTransZ += 0.1;
      }
      else if (g_chairTransZ >= -2){
    	  g_chairPulled = true;
      }

      // Can turn off internet if close enough AND power is on AND chair is pulled
      if (g_powerIsOn && g_chairPulled && !g_laptopClosed && g_distToLaptop > 1.0 && g_distToLaptop < 2.25 && g_laptopAngle > 0) {
    	  g_laptopAngle -= 5;
      }
      else if (g_laptopAngle <= 0) {
    	  g_laptopClosed = true;
      }

      // Can phone for help if close enough AND looked up rescue hotline on laptop
      if (g_laptopClosed && !g_phonedForHelp && g_distToPhone > 1.0 && g_distToPhone < 2.25) {
    	  g_phonedForHelp = true;
      }

      if (!g_hasKey && g_phonedForHelp && g_distToPlane > 1.0 && g_distToPlane < 2.25) {
    	  g_hasKey = true;
      }

      // Can open door if close enough AND has key from phoning rescue hotline
      if (g_hasKey && !g_doorOpened && g_distToDoor > 1.0 && g_distToDoor < 2.25  && g_doorAngle < 130.0) {
    	  g_doorAngle += 5;
      }
      else if (g_doorAngle >= 130.0){
    	  g_doorOpened = 1;
    	  g_hasKey = false;
      }

      break;
//    case 'p': // press to call for help
//      if (!g_phonedForHelp) {
//          g_phonedForHelp = true;
//      }
//      break;
//    case 'u':
//    	lightY += 0.1;
//    	break;
//    case 'n':
//    	lightY -= 0.1;
//    	break;
//    case 'h':
//    	lightX -= 0.1;
//    	break;
//    case 'j':
//    	lightX += 0.1;
//    	break;
  }
  glutPostRedisplay();
}

//routines to load in a bmp files - must be 2^nx2^m and a 24bit bmp
GLvoid LoadTexture(char* image_file, int texID) {

	TextureImage = (Image *) malloc(sizeof(Image));
	if (TextureImage == NULL) {
		printf("Error allocating space for image");
		exit(1);
	}
//	cout << "trying to load " << image_file << endl;
	if (!ImageLoad(image_file, TextureImage)) {
		exit(1);
	}
	/*  2d texture, level of detail 0 (normal), 3 components (red, green, blue),            */
	/*  x size from image, y size from image,                                              */
	/*  border 0 (normal), rgb color data, unsigned byte data, data  */
	glBindTexture(GL_TEXTURE_2D, texID);
	glTexImage2D(GL_TEXTURE_2D, 0, 3,
		TextureImage->sizeX, TextureImage->sizeY,
		0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage->data);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); /*  cheap scaling when image bigger than texture */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); /*  cheap scaling when image smalled than texture*/

}


/* BMP file loader loads a 24-bit bmp file only */

/*
* getint and getshort are help functions to load the bitmap byte by byte
*/
static unsigned int getint(FILE *fp) {
	int c, c1, c2, c3;

	/*  get 4 bytes */
	c = getc(fp);
	c1 = getc(fp);
	c2 = getc(fp);
	c3 = getc(fp);

	return ((unsigned int) c) +
		(((unsigned int) c1) << 8) +
		(((unsigned int) c2) << 16) +
		(((unsigned int) c3) << 24);
}

static unsigned int getshort(FILE *fp){
	int c, c1;

	/* get 2 bytes*/
	c = getc(fp);
	c1 = getc(fp);

	return ((unsigned int) c) + (((unsigned int) c1) << 8);
}

/*  quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only.  */

int ImageLoad(char *filename, Image *image) {
	FILE *file;
	unsigned long size;                 /*  size of the image in bytes. */
	unsigned long i;                    /*  standard counter. */
	unsigned short int planes;          /*  number of planes in image (must be 1)  */
	unsigned short int bpp;             /*  number of bits per pixel (must be 24) */
	char temp;                          /*  used to convert bgr to rgb color. */

	/*  make sure the file is there. */
	if ((file = fopen(filename, "rb"))==NULL) {
		printf("File Not Found : %s\n",filename);
		return 0;
	}

	/*  seek through the bmp header, up to the width height: */
	fseek(file, 18, SEEK_CUR);

	/*  No 100% errorchecking anymore!!! */

	/*  read the width */    image->sizeX = getint (file);

	/*  read the height */
	image->sizeY = getint (file);

	/*  calculate the size (assuming 24 bits or 3 bytes per pixel). */
	size = image->sizeX * image->sizeY * 3;

	/*  read the planes */
	planes = getshort(file);
	if (planes != 1) {
		printf("Planes from %s is not 1: %u\n", filename, planes);
		return 0;
	}

	/*  read the bpp */
	bpp = getshort(file);
	if (bpp != 24) {
		printf("Bpp from %s is not 24: %u\n", filename, bpp);
		return 0;
	}

	/*  seek past the rest of the bitmap header. */
	fseek(file, 24, SEEK_CUR);

	/*  read the data.  */
	image->data = (char *) malloc(size);
	if (image->data == NULL) {
		printf("Error allocating memory for color-corrected image data");
		return 0;
	}

	if ((i = fread(image->data, size, 1, file)) != 1) {
		printf("Error reading image data from %s.\n", filename);
		return 0;
	}

	for (i=0;i<size;i+=3) { /*  reverse all of the colors. (bgr -> rgb) */
		temp = image->data[i];
		image->data[i] = image->data[i+2];
		image->data[i+2] = temp;
	}

	fclose(file); /* Close the file and release the filedes */

	/*  we're done. */
	return 1;
}


int main(int argc, char *argv[]) {
	g_width = g_width = 800;

  glutInit(&argc, argv);
  glutInitWindowPosition(200, 200);
  glutInitWindowSize(g_width, g_width);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutCreateWindow("Final Project: Escape the Room");
  glutReshapeFunc(ReshapeGL);
  glutDisplayFunc(Draw);
  glutKeyboardFunc(keyboard);
  glutMouseFunc(mouseHandler);
  glutMotionFunc(motionHandler);
  glutIdleFunc(animation);

  LoadTexture((char *)"roomAlpha.bmp", 0);
  LoadTexture((char *)"hardwood.bmp", 1);
  LoadTexture((char *)"space.bmp", 2);


  InitGeom();

  // cube for room walls
  Meshes[0] = GeometryCreator::CreateRoomCube(vec3(12, 8, 12));

  // cubes for table
  Meshes[1] = GeometryCreator::CreateCube(vec3(3, 0.5, 3)); // tabletop
  Meshes[2] = GeometryCreator::CreateCube(vec3(0.25, 2, 0.25)); // table leg

  // cube and sphere for door
  Meshes[3] = GeometryCreator::CreateCube(vec3(1.75, 3.5, 0.25)); // door panel
  Meshes[4] = GeometryCreator::CreateSphere(vec3(0.1, 0.1, 0.1), 10, 10); // door knob

  // cube for chair
  Meshes[5] = GeometryCreator::CreateCube(vec3(1, 0.1, 1)); // chair seat
  Meshes[6] = GeometryCreator::CreateCube(vec3(1, 1, 0.10)); // chair backing

  // cube and cylinder for phone speaker and mouthpiece
  Meshes[7] = GeometryCreator::CreateCube(vec3(0.15, 0.6, 0.1));
  Meshes[8] = GeometryCreator::CreateCylinder(0.1f, 0.1f, 0.1f, 10, 10);


  #ifdef _WIN32
    GLenum err = glewInit();
    if (GLEW_OK != err) {
      cerr << "Error initializing glew! " << glewGetErrorString(err) << endl;
      return 1;
    }
  #endif
  Initialize();
  getGLversion();
  if (!InstallShader(textFileRead((char *)"prog4_Vert.glsl"), textFileRead((char *)"prog4_Frag.glsl"))) {
    printf("Error installing shader!\n");
    return 0;
  }


  glutMainLoop();
  return 0;
}





