struct Material {
  vec3 aColor;
  vec3 dColor;
  vec3 sColor;
  float shine;
};

uniform Material uMat;

uniform vec3 uLightColor;

varying vec3 vColor;
varying vec4 normal;
varying vec4 light;
varying vec4 view;

uniform sampler2D uTexUnit;
varying vec2 vTexCoord;

uniform int texFlag;
uniform int roomFlag;
uniform int texOnlyFlag;
uniform int gameWinFlag;

void main() {
  vec3 Diffuse, Spec, Ambient;
  float specValue;
  vec4 lightVector = normalize(light);
  vec4 norm = normalize(normal);
  
  // TEXTURE STUFF
  vec4 texColor0 = vec4(vColor.x, vColor.y, vColor.z, 1);
  vec4 texColor1 = texture2D(uTexUnit, vTexCoord);
  
  if (texOnlyFlag == 1) {
     gl_FragColor = vec4(texColor1[0], texColor1[1], texColor1[2], 1);
  }
  else if (texFlag == 1) {
     if (roomFlag == 1 && texColor1 == vec4(0, 0, 0, 1)) {
        discard;
     }
     
     Diffuse = uLightColor * max(0.0, dot(lightVector, norm)) * vec3(texColor1[0], texColor1[1], texColor1[2]);

     Ambient = uLightColor * vec3(0.1, 0.1, 0.1);

     Diffuse += Ambient;

     specValue = dot(normalize(view), -lightVector + 2.0 * dot(lightVector, norm) * norm);
  
     //specValue = (dot(normalize(view), reflect(-lightVector, norm)) * norm);
     Spec = uLightColor * clamp(pow(specValue, uMat.shine), 0.0, 1.0) * uMat.sColor;

     Diffuse += Spec;
     
     if (gameWinFlag == 1) {
        gl_FragColor = vec4(Ambient, 1.0);
     }
     else {
        gl_FragColor = vec4(Diffuse, 1.0);
     }  
  }
  else {

     Diffuse = uLightColor * max(0.0, dot(lightVector, norm)) * uMat.dColor;

     Ambient = uLightColor * uMat.aColor;

     Diffuse += Ambient;

     //specValue = dot(normalize(view), -lightVector + 2.0 * dot(lightVector, norm) * norm);
  
     specValue = (dot(normalize(view), reflect(-lightVector, norm)) * norm);
     Spec = uLightColor * clamp(pow(specValue, uMat.shine), 0.0, 1.0) * uMat.sColor;

     Diffuse += Spec;

     if (gameWinFlag == 1) {
        gl_FragColor = vec4(Ambient, 1.0);
     }
     else {
        gl_FragColor = vec4(Diffuse, 1.0);
     }  
     //gl_FragColor = normal;
  
  }
}
